@@ -1,15 +1,13 @@
 """Materialized Path Trees"""
 
-import sys
 import operator
-
-if sys.version_info >= (3, 0):
-    from functools import reduce
+from functools import reduce
 
 from django.core import serializers
 from django.db import models, transaction, connection
-from django.db.models import F, Q
-from django.utils.translation import ugettext_noop as _
+from django.db.models import F, Q, Value
+from django.db.models.functions import Concat, Substr
+from django.utils.translation import gettext_noop as _
 
 from treebeard.numconv import NumConv
 from treebeard.models import Node
@@ -274,15 +272,7 @@ class MP_ComplexAddMoveHandler(MP_AddHandler):
             connection.ops.quote_name(
                 get_result_class(self.node_cls)._meta.db_table), )
 
-        # <3 "standard" sql
-        if vendor == 'sqlite':
-            # I know that the third argument in SUBSTR (LENGTH(path)) is
-            # awful, but sqlite fails without it:
-            # OperationalError: wrong number of arguments to function substr()
-            # even when the documentation says that 2 arguments are valid:
-            # http://www.sqlite.org/lang_corefunc.html
-            sqlpath = "%s||SUBSTR(path, %s, LENGTH(path))"
-        elif vendor == 'mysql':
+        if vendor == 'mysql':
             # hooray for mysql ignoring standards in their default
             # configuration!
             # to make || work as it should, enable ansi mode
@@ -308,7 +298,7 @@ class MP_ComplexAddMoveHandler(MP_AddHandler):
 
 class MP_AddRootHandler(MP_AddHandler):
     def __init__(self, cls, **kwargs):
-        super(MP_AddRootHandler, self).__init__()
+        super().__init__()
         self.cls = cls
         self.kwargs = kwargs
 
@@ -332,7 +322,7 @@ class MP_AddRootHandler(MP_AddHandler):
         if len(self.kwargs) == 1 and 'instance' in self.kwargs:
             # adding the passed (unsaved) instance to the tree
             newobj = self.kwargs['instance']
-            if newobj.pk:
+            if not newobj._state.adding:
                 raise NodeAlreadySaved("Attempted to add a tree node that is "\
                     "already in the database")
         else:
@@ -348,7 +338,7 @@ class MP_AddRootHandler(MP_AddHandler):
 
 class MP_AddChildHandler(MP_AddHandler):
     def __init__(self, node, **kwargs):
-        super(MP_AddChildHandler, self).__init__()
+        super().__init__()
         self.node = node
         self.node_cls = node.__class__
         self.kwargs = kwargs
@@ -364,7 +354,7 @@ class MP_AddChildHandler(MP_AddHandler):
         if len(self.kwargs) == 1 and 'instance' in self.kwargs:
             # adding the passed (unsaved) instance to the tree
             newobj = self.kwargs['instance']
-            if newobj.pk:
+            if not newobj._state.adding:
                 raise NodeAlreadySaved("Attempted to add a tree node that is "\
                     "already in the database")
         else:
@@ -385,21 +375,23 @@ class MP_AddChildHandler(MP_AddHandler):
         else:
             # adding the new child as the last one
             newobj.path = self.node.get_last_child()._inc_path()
-        # saving the instance before returning it
-        newobj.save()
-        newobj._cached_parent_obj = self.node
 
         get_result_class(self.node_cls).objects.filter(
             path=self.node.path).update(numchild=F('numchild')+1)
 
         # we increase the numchild value of the object in memory
         self.node.numchild += 1
+
+        # saving the instance before returning it
+        newobj._cached_parent_obj = self.node
+        newobj.save()
+
         return newobj
 
 
 class MP_AddSiblingHandler(MP_ComplexAddMoveHandler):
     def __init__(self, node, pos, **kwargs):
-        super(MP_AddSiblingHandler, self).__init__()
+        super().__init__()
         self.node = node
         self.node_cls = node.__class__
         self.pos = pos
@@ -411,7 +403,7 @@ class MP_AddSiblingHandler(MP_ComplexAddMoveHandler):
         if len(self.kwargs) == 1 and 'instance' in self.kwargs:
             # adding the passed (unsaved) instance to the tree
             newobj = self.kwargs['instance']
-            if newobj.pk:
+            if not newobj._state.adding:
                 raise NodeAlreadySaved("Attempted to add a tree node that is "\
                     "already in the database")
         else:
@@ -452,7 +444,7 @@ class MP_AddSiblingHandler(MP_ComplexAddMoveHandler):
 
 class MP_MoveHandler(MP_ComplexAddMoveHandler):
     def __init__(self, node, target, pos=None):
-        super(MP_MoveHandler, self).__init__()
+        super().__init__()
         self.node = node
         self.node_cls = node.__class__
         self.target = target
@@ -637,6 +629,7 @@ class MP_Node(Node):
         if parent:
             qset = qset.filter(path__startswith=parent.path)
         ret, lnk = [], {}
+        pk_field = cls._meta.pk.attname
         for pyobj in serializers.serialize('python', qset):
             # django's serializer stores the attributes in 'fields'
             fields = pyobj['fields']
@@ -646,13 +639,13 @@ class MP_Node(Node):
             del fields['depth']
             del fields['path']
             del fields['numchild']
-            if 'id' in fields:
+            if pk_field in fields:
                 # this happens immediately after a load_bulk
-                del fields['id']
+                del fields[pk_field]
 
             newobj = {'data': fields}
             if keep_ids:
-                newobj['id'] = pyobj['pk']
+                newobj[pk_field] = pyobj['pk']
 
             if (not parent and depth == 1) or\
                (parent and len(path) == len(parent.path)):
@@ -727,7 +720,7 @@ class MP_Node(Node):
         return evil_chars, bad_steplen, orphans, wrong_depth, wrong_numchild
 
     @classmethod
-    def fix_tree(cls, destructive=False):
+    def fix_tree(cls, destructive=False, fix_paths=False):
         """
         Solves some problems that can appear when transactions are not used and
         a piece of code breaks, leaving the tree in an inconsistent state.
@@ -743,79 +736,147 @@ class MP_Node(Node):
               ``node_order_by`` is modified after the node is inserted, the
               tree ordering will be inconsistent.
 
-        :param destructive:
+        :param fix_paths:
 
-            A boolean value. If True, a more agressive fix_tree method will be
-            attempted. If False (the default), it will use a safe (and fast!)
-            fix approach, but it will only solve the ``depth`` and
+            A boolean value. If True, a slower, more complex fix_tree method
+            will be attempted. If False (the default), it will use a safe (and
+            fast!) fix approach, but it will only solve the ``depth`` and
             ``numchild`` nodes, it won't fix the tree holes or broken path
             ordering.
 
-            .. warning::
+        :param destructive:
 
-               Currently what the ``destructive`` method does is:
-
-               1. Backup the tree with :meth:`dump_data`
-               2. Remove all nodes in the tree.
-               3. Restore the tree with :meth:`load_data`
-
-               So, even when the primary keys of your nodes will be preserved,
-               this method isn't foreign-key friendly. That needs complex
-               in-place tree reordering, not available at the moment (hint:
-               patches are welcome).
+            Deprecated; alias for ``fix_paths``.
         """
         cls = get_result_class(cls)
         vendor = cls.get_database_vendor('write')
 
-        if destructive:
-            dump = cls.dump_bulk(None, True)
-            cls.objects.all().delete()
-            cls.load_bulk(dump, None, True)
-        else:
-            cursor = cls._get_database_cursor('write')
+        cursor = cls._get_database_cursor('write')
 
-            # fix the depth field
-            # we need the WHERE to speed up postgres
+        # fix the depth field
+        # we need the WHERE to speed up postgres
+        sql = (
+            "UPDATE %s "
+            "SET depth=" + sql_length("path", vendor=vendor) + "/%%s "
+            "WHERE depth!=" + sql_length("path", vendor=vendor) + "/%%s"
+        ) % (connection.ops.quote_name(cls._meta.db_table), )
+        vals = [cls.steplen, cls.steplen]
+        cursor.execute(sql, vals)
+
+        # fix the numchild field
+        vals = ['_' * cls.steplen]
+        # the cake and sql portability are a lie
+        if cls.get_database_vendor('read') == 'mysql':
             sql = (
-                "UPDATE %s "
-                "SET depth=" + sql_length("path", vendor=vendor) + "/%%s "
-                "WHERE depth!=" + sql_length("path", vendor=vendor) + "/%%s"
-            ) % (connection.ops.quote_name(cls._meta.db_table), )
-            vals = [cls.steplen, cls.steplen]
-            cursor.execute(sql, vals)
-
-            # fix the numchild field
-            vals = ['_' * cls.steplen]
-            # the cake and sql portability are a lie
-            if cls.get_database_vendor('read') == 'mysql':
-                sql = (
-                    "SELECT tbn1.path, tbn1.numchild, ("
-                    "SELECT COUNT(1) "
-                    "FROM %(table)s AS tbn2 "
-                    "WHERE tbn2.path LIKE " +
-                    sql_concat("tbn1.path", "%%s", vendor=vendor) + ") AS real_numchild "
-                    "FROM %(table)s AS tbn1 "
-                    "HAVING tbn1.numchild != real_numchild"
-                ) % {'table': connection.ops.quote_name(cls._meta.db_table)}
-            else:
-                subquery = "(SELECT COUNT(1) FROM %(table)s AS tbn2"\
-                           " WHERE tbn2.path LIKE " + sql_concat("tbn1.path", "%%s", vendor=vendor) + ")"
-                sql = ("SELECT tbn1.path, tbn1.numchild, " + subquery +
-                       " FROM %(table)s AS tbn1 WHERE tbn1.numchild != " +
-                       subquery)
-                sql = sql % {
+                "SELECT tbn1.path, tbn1.numchild, ("
+                "SELECT COUNT(1) "
+                "FROM %(table)s AS tbn2 "
+                "WHERE tbn2.path LIKE " +
+                sql_concat("tbn1.path", "%%s", vendor=vendor) + ") AS real_numchild "
+                "FROM %(table)s AS tbn1 "
+                "HAVING tbn1.numchild != real_numchild"
+            ) % {'table': connection.ops.quote_name(cls._meta.db_table)}
+        else:
+            subquery = "(SELECT COUNT(1) FROM %(table)s AS tbn2"\
+                        " WHERE tbn2.path LIKE " + sql_concat("tbn1.path", "%%s", vendor=vendor) + ")"
+            sql = ("SELECT tbn1.path, tbn1.numchild, " + subquery +
+                    " FROM %(table)s AS tbn1 WHERE tbn1.numchild != " +
+                    subquery)
+            sql = sql % {
+                'table': connection.ops.quote_name(cls._meta.db_table)}
+            # we include the subquery twice
+            vals *= 2
+        cursor.execute(sql, vals)
+        sql = "UPDATE %(table)s "\
+                "SET numchild=%%s "\
+                "WHERE path=%%s" % {
                     'table': connection.ops.quote_name(cls._meta.db_table)}
-                # we include the subquery twice
-                vals *= 2
+        for node_data in cursor.fetchall():
+            vals = [node_data[2], node_data[0]]
             cursor.execute(sql, vals)
-            sql = "UPDATE %(table)s "\
-                  "SET numchild=%%s "\
-                  "WHERE path=%%s" % {
-                      'table': connection.ops.quote_name(cls._meta.db_table)}
-            for node_data in cursor.fetchall():
-                vals = [node_data[2], node_data[0]]
-                cursor.execute(sql, vals)
 
+        if fix_paths or destructive:
+            with transaction.atomic():
+                # To fix holes and mis-orderings in paths, we consider each non-leaf node in turn
+                # and ensure that its children's path values are consecutive (and in the order
+                # given by node_order_by, if applicable). children_to_fix is a queue of child sets
+                # that we know about but have not yet fixed, expressed as a tuple of
+                # (parent_path, depth). Since we're updating paths as we go, we must take care to
+                # only add items to this list after the corresponding parent node has been fixed
+                # (and is thus not going to change).
+
+                # Initially children_to_fix is the set of root nodes, i.e. ones with a path
+                # starting with '' and depth 1.
+                children_to_fix = [('', 1)]
+
+                while children_to_fix:
+                    parent_path, depth = children_to_fix.pop(0)
+
+                    children = cls.objects.filter(
+                        path__startswith=parent_path, depth=depth
+                    ).values('pk', 'path', 'depth', 'numchild')
+
+                    desired_sequence = children.order_by(*(cls.node_order_by or ['path']))
+
+                    # mapping of current path position (converted to numeric) to item
+                    actual_sequence = {}
+
+                    # highest numeric path position currently in use
+                    max_position = None
+
+                    # loop over items to populate actual_sequence and max_position
+                    for item in desired_sequence:
+                        actual_position = cls._str2int(item['path'][-cls.steplen:])
+                        actual_sequence[actual_position] = item
+                        if max_position is None or actual_position > max_position:
+                            max_position = actual_position
+
+                    # loop over items to perform path adjustments
+                    for (i, item) in enumerate(desired_sequence):
+                        desired_position = i + 1  # positions are 1-indexed
+                        actual_position = cls._str2int(item['path'][-cls.steplen:])
+                        if actual_position == desired_position:
+                            pass
+                        else:
+                            # if a node is already in the desired position, move that node
+                            # to max_position + 1 to get it out of the way
+                            occupant = actual_sequence.get(desired_position)
+                            if occupant:
+                                old_path = occupant['path']
+                                max_position += 1
+                                new_path = cls._get_path(parent_path, depth, max_position)
+                                if len(new_path) > len(old_path):
+                                    previous_max_path = cls._get_path(parent_path, depth, max_position - 1)
+                                    raise PathOverflow(_("Path Overflow from: '%s'" % (previous_max_path, )))
+
+                                cls._rewrite_node_path(old_path, new_path)
+                                # update actual_sequence to reflect the new position
+                                actual_sequence[max_position] = occupant
+                                del(actual_sequence[desired_position])
+                                occupant['path'] = new_path
+
+                            # move item into the (now vacated) desired position
+                            old_path = item['path']
+                            new_path = cls._get_path(parent_path, depth, desired_position)
+                            cls._rewrite_node_path(old_path, new_path)
+                            # update actual_sequence to reflect the new position
+                            actual_sequence[desired_position] = item
+                            del(actual_sequence[actual_position])
+                            item['path'] = new_path
+
+                        if item['numchild']:
+                            # this item has children to process, and we have now moved the parent
+                            # node into its final position, so it's safe to add to children_to_fix
+                            children_to_fix.append((item['path'], depth + 1))
+
+    @classmethod
+    def _rewrite_node_path(cls, old_path, new_path):
+        cls.objects.filter(path__startswith=old_path).update(
+            path=Concat(
+                Value(new_path),
+                Substr('path', len(old_path) + 1)
+            )
+        )
 
     @classmethod
     def get_tree(cls, parent=None):
@@ -851,7 +912,7 @@ class MP_Node(Node):
         of *descendants* in every sibling.
         """
 
-        #~
+        # ~
         # disclaimer: this is the FOURTH implementation I wrote for this
         # function. I really tried to make it return a queryset, but doing so
         # with a *single* query isn't trivial with Django's ORM.
@@ -871,7 +932,7 @@ class MP_Node(Node):
         #
         # If there is a better way to do this in an UNMODIFIED django 1.0, let
         # me know.
-        #~
+        # ~
 
         cls = get_result_class(cls)
         vendor = cls.get_database_vendor('write')
@@ -894,7 +955,7 @@ class MP_Node(Node):
             '   COUNT(1)-1 AS count '
             '   FROM %(table)s '
             '   WHERE depth >= %(depth)s %(extrand)s'
-            '   GROUP BY '+ subpath + ') AS t2 '
+            '   GROUP BY ' + subpath + ') AS t2 '
             ' ON t1.path=t2.subpath '
             ' ORDER BY t1.path'
         ) % {
